// Slider.cpp: implementation of the CSlider class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Slider.h"
#include "TooltipService.h"
#include "main.h"
#include "utils.h"
#include "main.h"
#include "VolumeControl.h"

HWND g_hwndWinamp = NULL;

CSlider* GetSliderByWnd(HWND hWnd);

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CSlider::CSlider()
{
	m_bInitialized = false;
}

CSlider::~CSlider()
{
	if (m_nMode == MODE_WINAMP || m_nMode == MODE_WINAMPVOLUME)
	{
		KillTimer(m_hSlider, TIMER_UPDATE);
	}
	DestroyWindow(m_hHandle);
	DestroyWindow(m_hSlider);
}


bool CSlider::LoadParameters(char token[CONFIG_TOKENS_NUM][MAX_LINE_LENGTH], HBITMAP &hbmBackground, HBITMAP &hbmHandle)
{
	char *err;						// a pointer for strtod()

	// slider position
	m_nX = ParseCoordinate(token[1], 0, GetSystemMetrics(SM_CXSCREEN));
	m_nY = ParseCoordinate(token[2], 0, GetSystemMetrics(SM_CYSCREEN));

	// handle offset
	m_nHandleOffset = (int)strtod(token[3], &err);
	if (*err != 0)
	{
		utils::ErrorBoxPrint("Invalid definition of the slider:\n%s\nHandle offset must be an integer value", token[0]);
		return false;
	}

	// slider mode
	if (stricmp(token[7],"[VOLUME]") == 0)
	{
		if (*(token[8]) == 'W')
		{
			m_nMode = MODE_WINAMPVOLUME;
		}
		else
		{
			m_nMode = MODE_VOLUME;
			m_nChannel = (int)strtod(token[8], &err);
		}
	}
	else if (stricmp(token[7],"[BALANCE]") == 0)
	{
		m_nMode = MODE_BALANCE;
		m_nChannel = (int)strtod(token[8], &err);
	}
	else if (stricmp(token[7],"[WINAMP]") == 0)
	{
		m_nMode = MODE_WINAMP;
	}
	else
	{
		utils::ErrorBoxPrint("Invalid slider mode:\n%s\nProper values are [VOLUME], [BALANCE] and [WINAMP]", token[7]);
		return false;
	}

	// background and handle images
	hbmBackground = LoadLSImage(token[4], NULL);
	if (hbmBackground == NULL)
	{
		utils::ErrorBoxPrint("Cannot load a slider background image:\nFile %s not found or corrupt", token[4]);
		return false;
	}
	hbmHandle = LoadLSImage(token[5], NULL);
	if (hbmHandle == NULL)
	{
		utils::ErrorBoxPrint("Cannot load a slider handle image:\nFile %s not found or corrupt", token[5]);
		DeleteObject(hbmBackground);
		return false;
	}

	// flags
	m_bInverted	= (strchr(token[6], 'I') != NULL);
	m_bVisible	= (strchr(token[6], 'B') == NULL);
	m_nType = (strchr(token[6],'H') != NULL) ? TYPE_HORIZONTAL : TYPE_VERTICAL;
	char *pos;
	if ((pos = strchr(token[6], '#')) == NULL)
		m_nGroup = 0;
	else
		m_nGroup = (int)strtod(pos+1, &err);

	// calculate sizes
	GetLSBitmapSize(hbmBackground, &m_nWidth, &m_nHeight);
	GetLSBitmapSize(hbmHandle, &m_nHWidth, &m_nHHeight);

	m_nHX	= (m_nWidth - m_nHWidth)/2;
	m_nHY	= (m_nHeight - m_nHHeight)/2;

	return true;
}


//////////////////////////////////////////////////////////////////////////
//
// Create a slider window by filled
// slider structure and some other info
//
bool CSlider::CreateUI(
	const char *szCaption,
	HBITMAP background,
	HBITMAP handle,
	const char *flags)
{
	m_hSlider = CreateWindowEx(WS_EX_TOOLWINDOW |
		((strchr(flags,'A') == 0) ? 0 : WS_EX_TOPMOST),
		g_szSliderClass,
		szCaption,
		WS_POPUP,
		m_nX,
		m_nY,
		m_nWidth,
		m_nHeight,
		utils::GetLitestepDesktop(),
		NULL,
		g_hInstance,
		this);

	if (m_hSlider == NULL)
	{
		utils::ErrorBox("Cannot create slider window!");
		return false;
	}

	SetWindowLong(m_hSlider, GWL_USERDATA, magicDWord);
	g_pTooltips->Add(m_hSlider, szCaption);
	if (m_nMode == MODE_WINAMP || m_nMode == MODE_WINAMPVOLUME)
	{
		SetTimer(m_hSlider, TIMER_UPDATE, 250, NULL);
	}

	bool bVisible = LoadHandlePosition(false);
	if (!bVisible)
	{
		m_nHX = 0;
		m_nHY = 0;
	}

	m_hHandle = CreateWindowEx(
		WS_EX_TOOLWINDOW,
		g_szHandleClass, NULL, WS_CHILD,
		m_nHX, m_nHY, m_nHWidth, m_nHHeight,
		m_hSlider, NULL,
		g_hInstance, this);

	if (m_hHandle == NULL)
	{
		utils::ErrorBox("Cannot create handle window for the slider");
		DestroyWindow(m_hSlider);
		return false;
	}

	ShowWindow(m_hSlider, m_bVisible ? SW_SHOW : SW_HIDE);
	ShowWindow(m_hHandle, bVisible ? SW_SHOW : SW_HIDE);
	UpdateWindow(m_hSlider);
	UpdateWindow(m_hHandle);

	if (strchr(flags, 'T') != NULL)
	{
		SetWindowRgn(m_hSlider,BitmapToRegion(background,RGB(255,0,255),0x101010,0,0), TRUE);
		SetWindowRgn(m_hHandle,BitmapToRegion(handle,RGB(255,0,255),0x101010,0,0), TRUE);
	}
	HDC dc = GetDC(m_hSlider);
	m_dcMain = CreateCompatibleDC(dc);
	ReleaseDC(m_hSlider, dc);
	dc = GetDC(m_hHandle);
	m_dcHand = CreateCompatibleDC(dc);
	ReleaseDC(m_hHandle, dc);
	SelectObject(m_dcMain, background);
	SelectObject(m_dcHand, handle);
	DeleteObject(background);
	DeleteObject(handle);
	InvalidateRgn(m_hSlider, NULL, TRUE);
	PostMessage(m_hSlider, WM_UPDATE, 0, 0);
	return true;
}


///////////////////////////////////////////////////////////////////////////////////////
//
// This procedure processes all slider messages
//
LRESULT WINAPI SliderWindowProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	CSlider *pSlider = NULL;

	if (uMsg == WM_NCCREATE)
	{
		pSlider = reinterpret_cast<CSlider*>((reinterpret_cast<LPCREATESTRUCT>(lParam))->lpCreateParams);
		pSlider->m_hSlider = hWnd;
	}
	else
	{
		pSlider = GetSliderByWnd(hWnd);
	}

	if (pSlider == NULL)
	{
		return DefWindowProc(hWnd, uMsg, wParam, lParam);
	}

	switch(uMsg)
	{
	case WM_ERASEBKGND:
		BitBlt(
			(HDC)wParam,
			0,0,
			pSlider->m_nWidth,
			pSlider->m_nHeight,
			pSlider->m_dcMain,
			0,0,
			SRCCOPY);
		break;
	case WM_CHANGE:
		switch(pSlider->m_nMode)
		{
		case MODE_VOLUME:
			g_pMixer->ChangeVolume(pSlider->m_nChannel, wParam);
			break;
		case MODE_BALANCE:
			g_pMixer->ChangeBalance(pSlider->m_nChannel, ((100-wParam)*2)-100);
			break;
		case MODE_WINAMP:
			PostMessage(g_hwndWinamp, WM_USER, ((SendMessage(g_hwndWinamp,WM_USER, 1, 105)*1000)/100)*wParam, 106);
			break;
		case MODE_WINAMPVOLUME:
			PostMessage(g_hwndWinamp, WM_USER, wParam * 255 / 100, 122);
			break;
		}
		break;
	case WM_TIMER:
	case WM_UPDATE:
		pSlider->onUpdate();
		break;
	case WM_LBUTTONDOWN:
		pSlider->onLButtonDown(LOWORD(lParam), HIWORD(lParam));
		break;
	}
	return DefWindowProc(hWnd, uMsg, wParam, lParam);
}


///////////////////////////////////////////////////////////////////////////////////////
//
// This procedure processes all slider's handle messages
//
LRESULT WINAPI HandleWindowProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	CSlider *pSlider = NULL;

	if (uMsg == WM_NCCREATE)
	{
		pSlider = reinterpret_cast<CSlider*>((reinterpret_cast<LPCREATESTRUCT>(lParam))->lpCreateParams);
		pSlider->m_hHandle = hWnd;
	}
	else
	{
		pSlider = GetSliderByWnd(hWnd);
	}

	if (pSlider != NULL)
	switch(uMsg)
	{
	case WM_PAINT:
		{
			PAINTSTRUCT ps;
			HDC dc = BeginPaint(hWnd, &ps);
			BitBlt(
				dc,
				0, 0,
				pSlider->m_nHWidth,
				pSlider->m_nHHeight,
				pSlider->m_dcHand,
				0,0,
				SRCCOPY);
			EndPaint(hWnd,&ps);
		}
		break;
	case WM_NCHITTEST:
		return HTCAPTION;
	case WM_WINDOWPOSCHANGING:
		pSlider->onPosChanging(reinterpret_cast<WINDOWPOS*>(lParam));
		return 0;
	}
	return DefWindowProc(hWnd, uMsg, wParam, lParam);
}


void CSlider::onUpdate()
{
	m_bInitialized = true;
	if (LoadHandlePosition(true))
	{
		SetWindowPos(m_hHandle,0, m_nHX, m_nHY,0,0,SWP_NOSIZE | SWP_NOZORDER);
	}
}


void CSlider::onLButtonDown(int x, int y)
{
	if (m_nType == TYPE_HORIZONTAL)
	{
		if (m_nHandleOffset <= x && x <= m_nWidth - m_nHandleOffset)
		{
			SetWindowPos(m_hHandle, NULL, x, 0, 0, 0, SWP_NOSIZE | SWP_NOZORDER);
		}
	}
	else // if (m_nType == TYPE_VERTICAL)
	{
		if (m_nHandleOffset <= y && y <= m_nHeight - m_nHandleOffset)
		{
			SetWindowPos(m_hHandle, NULL, 0, y, 0, 0, SWP_NOSIZE | SWP_NOZORDER);
		}
	}
}


void CSlider::onPosChanging(WINDOWPOS *pPos)
{
	if (!m_bInitialized)
	{
		return;
	}
	if (m_nType == TYPE_HORIZONTAL)
	{
		pPos->x = max(pPos->x, m_nHandleOffset);
		pPos->x = min(pPos->x, m_nWidth - m_nHandleOffset*2 - m_nHWidth);
		pPos->y = 0;
		if (pPos->x != m_nHX)
		{
			m_nHX = pPos->x;
			int n = (m_nHX - m_nHandleOffset) * 100 / (m_nWidth - 2*m_nHandleOffset - m_nHWidth);
			if (!m_bInverted) n = 100 - n;
			PostMessage(m_hSlider, WM_CHANGE, n, 0);
		}
	}
	else	// if (m_nType == TYPE_VERTICAL)
	{
		pPos->y = max(pPos->y, m_nHandleOffset);
		pPos->y = min(pPos->y, m_nHeight - m_nHandleOffset*2 - m_nHHeight);
		pPos->x = 0;
		if (pPos->y != m_nHY)
		{
			m_nHY = pPos->y;
			int n = (m_nHY - m_nHandleOffset) * 100 / (m_nHeight - 2*m_nHandleOffset - m_nHHeight);
			if (!m_bInverted) n = 100 - n;
			PostMessage(m_hSlider, WM_CHANGE, n, 0);
		}
	}
}


bool CSlider::LoadHandlePosition(bool bExist)
{
	int left = 0, right = 0, tmp;
	switch (m_nMode)
	{
	case MODE_VOLUME:
		g_pMixer->GetVolume(m_nChannel, left, right);
		if (right>left)
			tmp = 100-(right/655);
		else
			tmp = 100-(left/655);
		break;
	case MODE_BALANCE:
		g_pMixer->GetVolume(m_nChannel,left,right);
		if (right > left)
			tmp = 100 - left/(right/100);
		else
			tmp = (right/(left/100)) - 100;
		tmp = (tmp+100)/2;
		break;
	case MODE_WINAMP:
	case MODE_WINAMPVOLUME:
		g_hwndWinamp = FindWindow("Winamp v1.x",NULL);
		if (g_hwndWinamp == NULL)
		{
			g_hwndWinamp = FindWindow("grdampclass",NULL);
		}
		if (g_hwndWinamp == NULL)
		{
			g_hwndWinamp = FindWindow("grdamp wndcls",NULL);
		}
		if (g_hwndWinamp == NULL)
		{
			if (bExist) ShowWindow(m_hHandle, SW_HIDE);
			return false;
		}
		else
		{
			if (bExist) ShowWindow(m_hHandle, SW_SHOW);
			int ret = SendMessage(g_hwndWinamp, WM_USER, 0, 104);
			if (ret != 1 && ret!=3)
			{
				if (bExist) ShowWindow(m_hHandle, SW_HIDE);
				return false;
			}
			else
			{
				if (bExist) ShowWindow(m_hHandle, SW_SHOW);
				if (m_nMode == MODE_WINAMPVOLUME)
				{
					tmp = SendMessage(g_hwndWinamp, WM_USER, 1, 122) * 100 / 255;
				}
				else // if (m_nMode == MODE_WINAMP)
				{
					tmp = SendMessage(g_hwndWinamp, WM_USER, 1, 105);
					tmp = 100 - 100 * SendMessage(g_hwndWinamp, WM_USER, 0, 105)/(tmp*1000);
				}
			}
		}
		break;
	}

	if (m_bInverted)
	{
		tmp = 100 - tmp;
	}
	if (m_nType == TYPE_VERTICAL)
	{
		m_nHY = (m_nHeight - 2 * m_nHandleOffset - m_nHHeight) * tmp / 100 + m_nHandleOffset;
	}
	else
	{
		m_nHX = (m_nWidth  - 2 * m_nHandleOffset - m_nHWidth) * tmp / 100 + m_nHandleOffset;
	}
	return true;
}


void CSlider::Show()
{
	if (!m_bVisible)
	{
		m_bVisible = true;
		ShowWindow(m_hSlider, SW_SHOW);
	}
}


void CSlider::Hide()
{
	if (m_bVisible)
	{
		m_bVisible = false;
		ShowWindow(m_hSlider, SW_HIDE);
	}
}


void CSlider::ToggleVisibility()
{
	m_bVisible = !m_bVisible;
	ShowWindow(m_hSlider, m_bVisible ? SW_SHOW : SW_HIDE);
}


void CSlider::Hook(HWND hwndParent)
{
	SetParent(m_hSlider, hwndParent);
}
