// umSlider.cpp : Defines the entry point for the DLL application.
//
#include "stdafx.h"
#include "main.h"
#include "utils.h"
#include "Slider.h"
#include "VolumeControl.h"
#include "TooltipService.h"

extern "C" {
    __declspec( dllexport ) int initModuleEx(HWND parent, HINSTANCE dll, LPCSTR szPath);
    __declspec( dllexport ) void quitModule(HINSTANCE dllInst);
}

HINSTANCE g_hInstance;

char g_szAppName[] = "UniMedia Slider";
char g_szSliderClass[] = "LSUMSlider";
char g_szHandleClass[] = "LSUMSliderHand";

int lsMessages[] =
{
	LM_GETREVID,
	0
};
HWND messageHandler;
LRESULT WINAPI MessageHandlerProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
const char rcsRevision[] = "0.2 /*for the internal use only*/";

CVolumeControl *g_pMixer = NULL;
CTooltipService *g_pTooltips = NULL;

typedef vector<CSlider*> SliderVector;
typedef vector<CSlider*>::iterator SliderIterator;

SliderVector sliders;
void InitSliders();
CSlider* ParseSliderConfig(const char *szLine);

LRESULT WINAPI SliderWindowProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT WINAPI HandleWindowProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);


///////////////////////////////////////////////////////////////////////////////
//
// Bangs...
// 
typedef struct BangItem
{
	char szName[25];
	BangCommand *pCommand;
} BangItem;

void bangVolumeUp		(HWND sender, LPCSTR args);
void bangVolumeDown		(HWND sender, LPCSTR args);
void bangMute			(HWND sender, LPCSTR args);
void bangBalanceLeft	(HWND sender, LPCSTR args);
void bangBalanceRight	(HWND sender, LPCSTR args);
void bangSetVolume		(HWND sender, LPCSTR args);
void bangSetBalance		(HWND sender, LPCSTR args);
void bangHideGroup		(HWND sender, LPCSTR args);
void bangShowGroup		(HWND sender, LPCSTR args);
void bangToggleGroup	(HWND sender, LPCSTR args);
void bangHook			(HWND sender, LPCSTR args);

// it is possible to end the list with something like
// {NULL,NULL}, but for what?
#define BANGS_NUM		14

BangItem g_Bangs[BANGS_NUM] =
{
	{"!SliderVolumeUp",		bangVolumeUp},
	{"!SliderVolumeDown",	bangVolumeDown},
	{"!SliderMute",			bangMute},
	{"!SliderBalanceLeft",	bangBalanceLeft},
	{"!SliderBalanceRight",	bangBalanceRight},
	{"!SliderSetVolume",	bangSetVolume},
	{"!SliderSetBalance",	bangSetBalance},
	{"!HideSliderGroup",	bangHideGroup},
	{"!SliderHideGroup",	bangHideGroup},
	{"!ShowSliderGroup",	bangShowGroup},
	{"!SliderShowGroup",	bangShowGroup},
	{"!ToggleSliderGroup",	bangToggleGroup},
	{"!SliderToggleGroup",	bangToggleGroup},
	{"!SliderHook",			bangHook}
};

void RegisterBangs();
void UnregisterBangs();

////////////////////////////////////////////////////////////////////
//
// Module entry point
//
int initModuleEx(HWND parent, HINSTANCE dll, LPCSTR szPath)
{
	g_hInstance = dll;

	// register the main window
	// it will handle LS and mixer messages

	WNDCLASSEX wc;
	wc.cbSize = sizeof(WNDCLASSEX);
	wc.style = CS_GLOBALCLASS;
	wc.lpfnWndProc = MessageHandlerProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = g_hInstance;
	wc.hbrBackground = NULL;
	wc.hCursor = NULL;
	wc.hIcon = NULL;
	wc.lpszMenuName = NULL;
	wc.lpszClassName = g_szAppName;
	wc.hIconSm = 0;

	if (!RegisterClassEx(&wc))
	{
		utils::ErrorBox("Cannot register main window class...");
		return 1;
	}

	ZeroMemory(&wc, sizeof(WNDCLASSEX));
	wc.cbSize = sizeof(WNDCLASSEX);
	wc.lpfnWndProc = SliderWindowProc;
	wc.lpszClassName = g_szSliderClass;
	wc.hInstance = g_hInstance;
	if (!RegisterClassEx(&wc))
	{
		utils::ErrorBox("Cannot register slider window class");
		return 1;
	}

	ZeroMemory(&wc, sizeof(WNDCLASSEX));
	wc.cbSize = sizeof(WNDCLASSEX);
	wc.lpfnWndProc = HandleWindowProc;
	wc.lpszClassName = g_szHandleClass;
	wc.hInstance = g_hInstance;
	if (!RegisterClassEx(&wc))
	{
		utils::ErrorBox("Cannot register handle window class");
		return 2;
	}

	messageHandler = CreateWindowEx(WS_EX_TOOLWINDOW,
		g_szAppName,
		0,
		WS_POPUP,
		0, 0, 0, 0, 
		0,
		0,
		dll,
		0);

	if (!messageHandler)
	{
		utils::ErrorBox("Cannot create main window...");
		return 1;
	}	
	SendMessage(GetLitestepWnd(), LM_REGISTERMESSAGE, (WPARAM) messageHandler, (LPARAM) lsMessages);

	// try to create a mixer
	g_pMixer = new CVolumeControl(messageHandler);
	if (g_pMixer->m_nCode != 0)
	{
		// delete mixer object to prevent bangs from failure
		delete g_pMixer;
		g_pMixer = NULL;
		// it is safe to exit here since quitModule check
		// all pointers
		return 1;
	}

	g_pTooltips = new CTooltipService();
	g_pTooltips->Init(g_hInstance);


	InitCommonControls();

	InitSliders();

	RegisterBangs();

	return 0;
}

///////////////////////////////////////////////////////////////////////////
//
// Destructor :)
//
void quitModule(HINSTANCE dllInst)
{
	UnregisterBangs();

	for (SliderIterator si = sliders.begin(); si != sliders.end(); si++)
	{
		delete *si;
	}
	sliders.clear();

	if (g_pMixer) delete g_pMixer;
	if (g_pTooltips) delete g_pTooltips;

	SendMessage(
		GetLitestepWnd(),
		LM_UNREGISTERMESSAGE,
		(WPARAM) messageHandler,
		(LPARAM) lsMessages);

	DestroyWindow(messageHandler);

	UnregisterClass(g_szAppName, dllInst);
	UnregisterClass(g_szSliderClass, dllInst);
	UnregisterClass(g_szHandleClass, dllInst);
}

////////////////////////////////////////////////////////////////////////////
//
// Main umSlider window created to process the following messages:
//	LM_GETREVID				-	Revision ID
//	MM_MIXM_LINE_CHANGE		-	Callback notification is much better
//	MM_MIXM_CONTROL_CHANGE		than timer
//
LRESULT WINAPI MessageHandlerProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch(uMsg)
	{
	case LM_GETREVID:
		LPSTR buf;
		buf = (LPSTR)lParam;
		switch (wParam)
		{
		case 0:
		case 1:
			StringCchPrintf(buf, lParam, "umSlider: %s (Seg@)", rcsRevision);
			break;
		default:
			StringCchCopy(buf, lParam, "");
		}
		return strlen(buf);
	case MM_MIXM_LINE_CHANGE:
	case MM_MIXM_CONTROL_CHANGE:
		for (SliderIterator si = sliders.begin(); si != sliders.end(); si++)
		{
			if ((*si)->m_nType != MODE_WINAMP && (*si)->m_nType != MODE_WINAMPVOLUME)
			{
				PostMessage((*si)->m_hSlider, WM_UPDATE, 0, 0);
			}
		}
		break;
	}
	return DefWindowProc(hWnd, uMsg, wParam, lParam);
}


///////////////////////////////////////////////////////////////////////////////////////
//
// Read the config and create sliders
//
void InitSliders()
{
	FILE *f = LCOpen(NULL);		// step.rc
	if (!f)
	{
		utils::ErrorBox("Cannot open step.rc for reading the configuration");
		return;
	}

	// let's enumerate sliders

	char szLine[MAX_LINE_LENGTH];	// current line

	while (LCReadNextConfig(f, "*Slider", szLine, MAX_LINE_LENGTH))
	{
		char *p = szLine + strlen("*Slider");
		while (*p == ' ' || *p == '\t') p++;
		ParseSliderConfig(p);
	}
	if (!LCClose(f))
	{
		utils::ErrorBox("Error closing .RC");
	}
}


CSlider* ParseSliderConfig(const char *szLine)
{
	HBITMAP background, handle;		// bitmaps :)
	CSlider *pSlider;				// temporary variable where slider info will be stored

	// there are 10 tokens in the config string
	// let's prepare string array
	char
		token[CONFIG_TOKENS_NUM][MAX_LINE_LENGTH],
		*tokens[CONFIG_TOKENS_NUM],
		szExtra[MAX_LINE_LENGTH];

	for (int i = 0; i < CONFIG_TOKENS_NUM; i++)
	{
		tokens[i] = token[i];
	}

	pSlider = new CSlider();
	LCTokenize(szLine, tokens, CONFIG_TOKENS_NUM, szExtra);
	if (!pSlider->LoadParameters(token, background, handle))
	{
		delete pSlider;
		pSlider = NULL;
	}
	else
	{
		sliders.push_back(pSlider);
		if (!pSlider->CreateUI(token[0], background, handle, token[6]))
		{
			sliders.erase(sliders.end() - 1);
			delete pSlider;
			pSlider = NULL;
		}
	}
	return pSlider;
}

///////////////////////////////////////////////////////////////////////////////
//
// Find slider by window handle (both main and handle)
//
CSlider* GetSliderByWnd(HWND hWnd)
{
	for (SliderVector::iterator si = sliders.begin(); si != sliders.end(); si ++)
	{
		if ((*si)->m_hSlider == hWnd || (*si)->m_hHandle == hWnd)
		{
			return (*si);
		}
	}
	return NULL;
}


///////////////////////////////////////////////////////////////////////////////
//
// Register !bang commands
//
void RegisterBangs()
{
	for (int i = 0; i < BANGS_NUM; i++)
	{
		AddBangCommand(g_Bangs[i].szName, g_Bangs[i].pCommand);
	}
}


///////////////////////////////////////////////////////////////////////////////
//
// Unregister them
//
void UnregisterBangs()
{
	for (int i = 0; i < BANGS_NUM; i++)
	{
		RemoveBangCommand(g_Bangs[i].szName);
	}
}

///////////////////////////////////////////////////////////////////////////////
//
// Volume control bangs
//

void bangVolumeUp(HWND sender, LPCSTR args)
{
	if (g_pMixer) g_pMixer->VolumeUp(atoi(args));
}


void bangVolumeDown(HWND sender, LPCSTR args)
{
	if (g_pMixer) g_pMixer->VolumeDown(atoi(args));
}

void bangMute(HWND sender, LPCSTR args)
{
	int channel = atoi(args);
	if (g_pMixer)
	{
		g_pMixer->SetMute(channel, g_pMixer->GetMute(channel));
	}
}

void bangBalanceLeft(HWND sender, LPCSTR args)
{
	if (g_pMixer) g_pMixer->BalanceLeft(atoi(args));
}

void bangBalanceRight(HWND sender, LPCSTR args)
{
	if (g_pMixer) g_pMixer->BalanceRight(atoi(args));
}

void bangSetBalance(HWND sender, LPCSTR args)
{
	char
		token1[MAX_LINE_LENGTH], token2[MAX_LINE_LENGTH],
		szExtra[MAX_LINE_LENGTH], *tokens[] = {token1, token2};
	token1[0] = token2[0] = 0;
	LCTokenize(args, tokens, 2, szExtra);
	if (g_pMixer) g_pMixer->ChangeBalance(atoi(token1), atoi(token2));
}

void bangSetVolume(HWND sender, LPCSTR args)
{
	char
		token1[MAX_LINE_LENGTH], token2[MAX_LINE_LENGTH],
		szExtra[MAX_LINE_LENGTH], *tokens[] = {token1, token2};
	token1[0] = token2[0] = 0;
	LCTokenize(args, tokens, 2, szExtra);
	if (g_pMixer) g_pMixer->ChangeVolume(atoi(token1), atoi(token2));
}

///////////////////////////////////////////////////////////////////////////////
//
// Slider bangs
//

void bangHideGroup(HWND sender, LPCSTR args)
{
	int nGroup = atoi(args);
	for (SliderIterator si = sliders.begin(); si != sliders.end(); si ++)
	{
		if ((*si)->m_nGroup == nGroup)
		{
			(*si)->Hide();
		}
	}
}

void bangShowGroup(HWND sender, LPCSTR args)
{
	int nGroup = atoi(args);
	for (SliderIterator si = sliders.begin(); si != sliders.end(); si ++)
	{
		if ((*si)->m_nGroup == nGroup)
		{
			(*si)->Show();
		}
	}
}

void bangToggleGroup(HWND sender, LPCSTR args)
{
	int nGroup = atoi(args);
	for (SliderIterator si = sliders.begin(); si != sliders.end(); si ++)
	{
		if ((*si)->m_nGroup == nGroup)
		{
			(*si)->ToggleVisibility();
		}
	}
}


///////////////////////////////////////////////////////////////////////////////
//
// Hook to LSBox or similar module
//

void bangHook(HWND sender, LPCSTR args)
{
	char *p;
	if ((p = strrchr(args, ' ')) != NULL)
	{
		HWND hwndParent = reinterpret_cast<HWND>(atoi(p + 1));
		if (hwndParent == NULL) return;
		char szConfig[MAX_LINE_LENGTH];
		StringCchCopy(szConfig, MAX_LINE_LENGTH, args);
		p = strrchr(szConfig, ' ');
		*p = 0;
		CSlider *pSlider = ParseSliderConfig(szConfig);
		if (pSlider)
		{
			pSlider->Hook(hwndParent);
		}
	}
}

